@@ -1,4 +1,8 @@
 from webob import Request
+try:
+    sorted
+except NameError:
+    from webtest import sorted
 
 __all__ = ['debug_app']
 
@@ -6,7 +10,7 @@ def debug_app(environ, start_response):
     req = Request(environ)
     if 'error' in req.GET:
         raise Exception('Exception requested')
-    status = req.GET.get('status', '200 OK')
+    status = str(req.GET.get('status', '200 OK'))
     parts = []
     for name, value in sorted(environ.items()):
         if name.upper() != name:
@@ -25,6 +29,8 @@ def debug_app(environ, start_response):
             header_name = name[len('header-'):]
             headers.append((header_name, value))
     start_response(status, headers)
+    if req.method == 'HEAD':
+        return ['']
     return [body]
 
 def make_debug_app(global_conf):
